"""
Database helper classes using Factory Pattern
"""
import logging
import os
from abc import (
    ABCMeta,
    abstractmethod,
)

from etlhelper.exceptions import ETLHelperConnectionError

logger = logging.getLogger('etlhelper')


class DbHelper(metaclass=ABCMeta):
    """
    Abstract Base Class for DBHelpers
    """
    sql_exceptions = None
    connect_exceptions = None
    table_info_query = None
    # The following are used to help create parameterized queries.  Although
    # paramstyle is required by DBAPI2, most drivers support both a named and
    # positional style.
    paramstyle = None
    named_paramstyle = None
    positional_paramstyle = None

    @abstractmethod
    def __init__(self):
        self.sql_exceptions = tuple()
        self.connect_exceptions = tuple()
        self.required_params = set()
        self.paramstyle = ''
        self.missing_driver_msg = ''
        # This is overridden with real connect method when DbHelper class is
        # successfully initialised if driver is installed
        self._connect_func = self._raise_missing_driver_error_on_connect

    def connect(self, db_params, password_variable=None, **kwargs):
        """
        Return a connection (as appropriate), configured for
        the database with the password obtained from environment variable.  These
        connection classes provide Python's dbapi interface (see PEP 249).  The
        dbapi interface interacts with data row-by-row and is used for low-level
        functions.

        :param db_params: DbParams object or similar with appropriate attributes
        :param password_variable: str, name of environment variable with password
        :param kwargs: connection specific keyword arguments e.g. encoding
        :return: Connection object
        """
        # Prepare connection string
        conn_str = self.get_connection_string(db_params, password_variable)

        # Create connection
        try:
            conn = self._connect_func(conn_str, **kwargs)
        except self.connect_exceptions as exc:
            msg = f"Error connecting to {db_params} via dbapi: {exc}"
            raise ETLHelperConnectionError(msg)
        return conn

    @staticmethod
    def get_password(password_variable):
        """
        Read password from environment variable.
        :param password_variable: str, name of environment variable with password
        :return: str, password
        :raises ETLHelperDbParamsError: Exception when parameter not defined
        """
        if not password_variable:
            msg = "Name of password environment variable e.g. ORACLE_PASSWORD is required"
            logger.error(msg)
            raise ETLHelperConnectionError(msg)
        try:
            return os.environ[password_variable]
        except KeyError:
            msg = f"Password environment variable ({password_variable}) is not set"
            logger.error(msg)
            raise ETLHelperConnectionError(msg)

    @staticmethod
    @abstractmethod
    def get_connection_string(db_params, password_variable):
        """
        :returns: str
        """
        return

    @staticmethod
    def executemany(cursor, query, chunk):
        """
        Call executemany method appropriate to database.  Overridden for PostgreSQL.

        :param cursor: Open database cursor.
        :param query: str, SQL query
        :param chunk: list, Rows of parameters.
        """
        cursor.executemany(query, chunk)

    @staticmethod
    def cursor(conn):
        """
        Return a cursor on the connection.  Overridded for SQLite.

        :param conn: Open database connection.
        """
        return conn.cursor()

    def _raise_missing_driver_error_on_connect(self, *args, **kwargs):
        """
        Raise an exception with helpful message if user tries to connect without driver installed.
        This function replaces a connect function, so *args and **kwargs are collected to allow
        it to accept whatever would be passed to that function.
        """
        raise ETLHelperConnectionError(self.missing_driver_msg)



"""ETL Helper script to demonstrate copy_rows."""
import etlhelper as etl
from etlhelper.row_factories import namedtuple_row_factory
from my_databases import POSTGRESDB, ORACLEDB

select_sql = """
    SELECT
        customer_id,
        SUM (amount) AS total_amount
    FROM payment
    WHERE id > 1000
    GROUP BY customer_id
"""

# This insert query uses positional parameters, so a namedtuple_row_factory
# is used.
insert_sql = """
    INSERT INTO dest (
        customer_id,
        total_amount,
        loaded_by,
        load_time
    )
    VALUES (
        %s,
        %s,
        current_user,
        now()
    )
"""

with ORACLEDB.connect("ORA_PASSWORD") as src_conn:
    with POSTGRESDB.connect("PG_PASSWORD") as dest_conn:
        etl.copy_rows(
            select_sql,
            src_conn,
            insert_sql,
            dest_conn,
            row_factory=namedtuple_row_factory,
        )



"""
This module defines the DbParams class for storing database connection
parameters.
"""
import os
import socket

from etlhelper.connect import (
    connect,
    get_connection_string,
    get_sqlalchemy_connection_string,
)
from etlhelper.db_helper_factory import DB_HELPER_FACTORY
from etlhelper.exceptions import (
    ETLHelperDbParamsError,
    ETLHelperHelperError,
)


class DbParams(dict):
    """Generic data holder class for database connection parameters."""

    """
    As we do not know which parameters will be provided in advance, DbParams
    subclasses dict, to give dynamic attributes, following the pattern described
    here: https://amir.rachum.com/blog/2016/10/05/python-dynamic-attributes/
    """
    def __init__(self, dbtype='dbtype not set', **kwargs):
        kwargs.update(dbtype=dbtype.upper())
        super().__init__(kwargs)
        self.validate_params()

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            # getattr should raise AttributeError, not KeyError
            # https://docs.python.org/3/library/functions.html#getattr
            raise AttributeError(f'No such attribute: {item}')

    def __setattr__(self, item, value):
        self[item] = value

        # Validate the updated parameter set, and remove bad item if failed
        try:
            self.validate_params()
        except ETLHelperDbParamsError:
            self.pop(item)
            raise

    def __dir__(self):
        return super().__dir__() + [str(k) for k in self.keys()]

    def validate_params(self):
        """
        Validate database parameters.

        Should validate that a dbtype is a valid one and that the appropriate
        params have been passed for a particular db_type.

        :raises ETLHelperParamsError: Error if params are invalid
        """
        # Get a set of the attributes to compare against required attributes.
        given = set(self.keys())

        try:
            required_params = DB_HELPER_FACTORY.from_dbtype(self.dbtype).required_params
        except ETLHelperHelperError:
            msg = f'{self.dbtype} not in valid types ({DB_HELPER_FACTORY.helpers.keys()})'
            # from None suppresses lower errors in the stack trace
            # Deeper error is recorded in ETL HelperDbParamsError.__context__
            raise ETLHelperDbParamsError(msg) from None

        unset_params = (given ^ required_params) & required_params
        if unset_params:
            msg = f'{unset_params} not set. Required parameters are {required_params}'
            raise ETLHelperDbParamsError(msg)

        valid_params = required_params.union({'dbtype'})
        bad_params = given ^ valid_params
        if bad_params:
            msg = f"Invalid parameter(s): {bad_params}"
            raise ETLHelperDbParamsError(msg)

    @classmethod
    def from_environment(cls, prefix='ETLHelper_'):
        """
        Create DbParams object from parameters specified by environment
        variables e.g. ETLHelper_dbtype, ETLHelper_host, ETLHelper_port, etc.
        :param prefix: str, prefix to environment variable names
        """
        dbparams_keys = [key for key in os.environ if key.startswith(prefix)]
        dbparams_from_env = {key.replace(prefix, '').lower(): os.environ[key]
                             for key in dbparams_keys}

        # Ensure dbtype has been set
        dbtype_var = f'{prefix}dbtype'
        dbtype = dbparams_from_env.get('dbtype', None)
        if dbtype is None:
            msg = f"{dbtype_var} environment variable is not set"
            raise ETLHelperDbParamsError(msg)

        # Only include the required params
        # This prevents something like ETLHelper_password being added
        required_params = DB_HELPER_FACTORY.from_dbtype(dbtype).required_params | {'dbtype'}
        dbparams_from_env = {key: dbparams_from_env[key] for key in required_params}

        return cls(**dbparams_from_env)

    def is_reachable(self):
        """
        Test whether network allows opening of tcp/ip connection to database. No
        username or password are required.

        :return bool:
        """
        items = dict(self.items())
        if items['dbtype'] == 'SQLITE':
            raise ValueError("SQLITE DbParam does not require connection over network")

        s = socket.socket()
        s.settimeout(5)
        try:
            # Connection succeeds
            s.connect((items['host'], int(items['port'])))
            return True
        except OSError:
            # Failed to connect
            return False
        finally:
            s.close()

    def connect(self, password_variable=None, **kwargs):
        """
        Return database connection.

        :param password_variable: str, name of environment variable with password
        :param kwargs: connection specific keyword arguments e.g. row_factory
        :return: Connection object
        """
        return connect(self, password_variable, **kwargs)

    def get_connection_string(self, password_variable=None):
        """
        Return the connection string.

        :param password_variable: str, name of environment variable with password
        :return: str, Connection string
        """
        return get_connection_string(self, password_variable)

    def get_sqlalchemy_connection_string(self, password_variable=None):
        """
        Get a SQLAlchemy connection string.

        :param password_variable: str, name of environment variable with password
        :return: str, Connection string
        """
        return get_sqlalchemy_connection_string(self, password_variable)

    def copy(self):
        """
        Return a shallow copy of DbParams object.

        :return DbParams: DbParams object with same attributes as original.
        """
        return self.__class__(**self)

    @property
    def paramstyle(self):
        """The DBAPI2 paramstyle attribute for database type"""
        return DB_HELPER_FACTORY.from_dbtype(self.dbtype).paramstyle

    def __repr__(self):
        key_val_str = ", ".join([f"{key}='{self[key]}'" for key in self.keys()])
        return f"DbParams({key_val_str})"

    def __str__(self):
        return self.__repr__()



"""ETL Helper script to load records to a database table."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"

create_sql = """
    CREATE TABLE IF NOT EXISTS igneous_rock (
        id INTEGER PRIMARY KEY,
        name TEXT UNIQUE,
        grain_size TEXT
    )
"""

igneous_rocks = [
    {"name": "basalt", "grain_size": "fine"},
    {"name": "granite", "grain_size": "coarse"}
]


with sqlite3.connect(db_file) as conn:
    # Create table
    etl.execute(create_sql, conn)

    # Insert rows
    etl.load("igneous_rock", conn, igneous_rocks)

    # Confirm selection
    for row in etl.fetchall("SELECT * FROM igneous_rock", conn):
        print(row)



"""ETL Helper script to demonstrate logging errors."""
import logging
import sqlite3
import etlhelper as etl

etl.log_to_console()
logger = logging.getLogger("etlhelper")

db_file = "igneous_rocks.db"

rows = [
    {"name": "basalt", "grain_size": "fine"},
    {"name": "granite", "grain_size": "coarse"}
]


def log_errors(failed_rows: list[tuple[dict, Exception]]) -> None:
    for row, exception in failed_rows:
        logger.error(exception)


with sqlite3.connect(db_file) as conn:
    etl.load("igneous_rock", conn, rows, on_error=log_errors)



"""
DbHelper classes are used behind-the-scenes for customising the behaviour of
different database drivers.
They are not normally called directly.

For more details, see the source code in each module.
"""
# flake8: noqa
from etlhelper.db_helpers.db_helper import DbHelper
from etlhelper.db_helpers.oracle import OracleDbHelper
from etlhelper.db_helpers.postgres import PostgresDbHelper
from etlhelper.db_helpers.mssql import MSSQLDbHelper
from etlhelper.db_helpers.sqlite import SQLiteDbHelper



"""ETL Helper script to demonstrate using fetch functions with a given row factory."""
import sqlite3
import etlhelper as etl
from etlhelper.row_factories import namedtuple_row_factory

db_file = "igneous_rocks.db"

with sqlite3.connect(db_file) as conn:
    row = etl.fetchone(
            "SELECT * FROM igneous_rock",
            conn,
            row_factory=namedtuple_row_factory,
        )

print(row)
print(row.name)



"""ETL Helper script to demonstrate using Apache Airflow to schedule tasks."""
import datetime as dt
from airflow import DAG
from airflow.operators.python_operator import PythonOperator
from database_to_database import copy_readings


def copy_readings_with_args(**kwargs) -> None:
    # Set arguments for copy_readings from context
    start = kwargs.get("prev_execution_date")
    end = kwargs.get("execution_date")
    copy_readings(start, end)


dag = DAG(
    "readings",
    schedule_interval=dt.timedelta(days=1),
    start_date=dt.datetime(2019, 8, 1),
    catchup=True,
)

t1 = PythonOperator(
    task_id="copy_readings",
    python_callable=copy_readings_with_args,
    provide_context=True,
    dag=dag,
)



"""
Fixtures for pytest.  Functions defined here can be passed as arguments to
pytest tests.  scope parameter describes how often they are recreated e.g.
once per module.
"""
import logging
import os
import socket
import datetime as dt
from collections import namedtuple
from pathlib import Path
from textwrap import dedent
from typing import Any
from zipfile import ZipFile

import pytest
from psycopg2.extras import execute_batch

from etlhelper import (
    connect,
    log_to_console,
    DbParams,
)

PGTESTDB = DbParams(
    dbtype='PG',
    host='localhost',
    port=5432,
    dbname='etlhelper',
    user='etlhelper_user')


@pytest.fixture(scope="function")
def logger() -> logging.Logger:
    """
    Return an enabled etlhelper logger for tests.
    The logger handler is set to NullHandler afterwards.
    """
    log_to_console()
    logger = logging.getLogger("etlhelper")
    yield logger
    logger.handlers.clear()


@pytest.fixture(scope='module')
def pgtestdb_insert_sql():
    """Return SQL command used to populate test database."""
    insert_sql = dedent("""
          INSERT INTO src (id, value, simple_text, utf8_text,
            day, date_time)
          VALUES
            (%s, %s, %s, %s, %s, %s)
            ;""").strip()
    return insert_sql


@pytest.fixture(scope='function')
def pgtestdb_test_tables(test_table_data_namedtuple, pgtestdb_conn, pgtestdb_insert_sql):
    """
    Create a table and fill with test data.  Teardown after the yield drops it
    again.
    """
    drop_src_sql = "DROP TABLE IF EXISTS src;"
    create_src_sql = dedent("""
          CREATE TABLE src
            (
              id integer primary key,
              value double precision not null,
              simple_text text default 'default',
              utf8_text text,
              day date,
              date_time timestamp without time zone
            )
            ;""").strip()
    drop_dest_sql = drop_src_sql.replace('src', 'dest')
    create_dest_sql = create_src_sql.replace('src', 'dest')

    # Create table and populate with test data
    with pgtestdb_conn.cursor() as cursor:
        cursor.execute(drop_src_sql)
        cursor.execute(drop_dest_sql)
        cursor.execute(create_src_sql)
        cursor.execute(create_dest_sql)
        execute_batch(cursor, pgtestdb_insert_sql, test_table_data_namedtuple)
    pgtestdb_conn.commit()

    # Return control to calling function until end of test
    yield

    # Tear down the table after test completes
    with pgtestdb_conn.cursor() as cursor:
        cursor.execute(drop_src_sql)
        cursor.execute(drop_dest_sql)
    pgtestdb_conn.commit()


@pytest.fixture(scope='module')
def test_table_data_dict() -> list[dict[str, Any]]:
    """
    Return list of dictionaries of test data
    """
    data = [
        {"id": 1, "value": 1.234, "simple_text": "text", "utf8_text": 'Öæ°\nz',
         "day": dt.date(2018, 12, 7), "date_time": dt.datetime(2018, 12, 7, 13, 1, 59)},
        {"id": 2, "value": 2.234, "simple_text": "text", "utf8_text": 'Öæ°\nz',
         "day": dt.date(2018, 12, 8), "date_time": dt.datetime(2018, 12, 8, 13, 1, 59)},
        {"id": 3, "value": 2.234, "simple_text": "text", "utf8_text": 'Öæ°\nz',
         "day": dt.date(2018, 12, 9), "date_time": dt.datetime(2018, 12, 9, 13, 1, 59)},
    ]
    return data


@pytest.fixture(scope='module')
def test_table_data_namedtuple():
    """
    Return list of tuples of test data
    """
    Row = namedtuple('Row',
                     'id, value, simple_text, utf8_text, day, date_time')
    data = [
        Row(1, 1.234, 'text', 'Öæ°\nz', dt.date(2018, 12, 7),
            dt.datetime(2018, 12, 7, 13, 1, 59)),
        Row(2, 2.234, 'text', 'Öæ°\nz', dt.date(2018, 12, 8),
            dt.datetime(2018, 12, 8, 13, 1, 59)),
        Row(3, 2.234, 'text', 'Öæ°\nz', dt.date(2018, 12, 9),
            dt.datetime(2018, 12, 9, 13, 1, 59)),
    ]
    return data


@pytest.fixture(scope='function')
def pgtestdb_conn(pgtestdb_params):
    """Get connection to test PostGIS database."""
    with connect(pgtestdb_params, 'TEST_PG_PASSWORD') as conn:
        yield conn


@pytest.fixture(scope='module')
def pgtestdb_params():
    """
    Create DbParams object for test PostGIS database. Override hostname from
    environment variable to allow CI pipeline to use its own database.
    """
    pg_test_db = PGTESTDB.copy()

    # Override host and port if defined in environment variable
    host = os.getenv('TEST_PG_HOST', None)
    if host:
        pg_test_db.update(host=host)

    port = os.getenv('TEST_PG_PORT', None)
    if port:
        pg_test_db.update(port=port)

    return pg_test_db


def db_is_unreachable(host, port):
    """
    Attempt to connect to generic host, port combination to check network.
    :param host:
    :param port:
    :return: boolean
    """
    s = socket.socket()
    try:
        # Connection succeeds
        s.connect((host, int(port)))
        return False
    except OSError:
        # Failed to connect
        return True
    finally:
        s.close()


@pytest.fixture()
def dummy_zipfile(tmp_path):
    """
    Return a mocked up zip file for testing setup_oracle_client.py
    """
    names = ['instantclient_19_6/adrci', 'instantclient_19_6/BASIC_LICENSE',
             'instantclient_19_6/BASIC_README', 'instantclient_19_6/genezi',
             'instantclient_19_6/libclntshcore.so.19.1', 'instantclient_19_6/libclntsh.so',
             'instantclient_19_6/libclntsh.so.10.1', 'instantclient_19_6/libclntsh.so.11.1',
             'instantclient_19_6/libclntsh.so.12.1', 'instantclient_19_6/libclntsh.so.18.1',
             'instantclient_19_6/libclntsh.so.19.1', 'instantclient_19_6/libipc1.so',
             'instantclient_19_6/libmql1.so', 'instantclient_19_6/libnnz19.so',
             'instantclient_19_6/libocci.so', 'instantclient_19_6/libocci.so.10.1',
             'instantclient_19_6/libocci.so.11.1', 'instantclient_19_6/libocci.so.12.1',
             'instantclient_19_6/libocci.so.18.1', 'instantclient_19_6/libocci.so.19.1',
             'instantclient_19_6/libociei.so', 'instantclient_19_6/libocijdbc19.so',
             'instantclient_19_6/liboramysql19.so', 'instantclient_19_6/network/',
             'instantclient_19_6/ojdbc8.jar', 'instantclient_19_6/ucp.jar',
             'instantclient_19_6/uidrvci', 'instantclient_19_6/xstreams.jar',
             'instantclient_19_6/network/admin/', 'instantclient_19_6/network/admin/README']

    dummy_zipfile = ZipFile(tmp_path / 'instantclient.zip', 'w')

    # Create files in the archive that just contain filename
    for name in names:
        dummy_zipfile.writestr(name, name)
    dummy_zipfile.close()

    return Path(dummy_zipfile.filename)



"""
Database helper for Oracle
"""
import warnings
from textwrap import dedent

from etlhelper.db_helpers.db_helper import DbHelper


class OracleDbHelper(DbHelper):
    """
    Oracle DB helper class
    """
    table_info_query = dedent("""
        SELECT
            column_name as name,
            data_type as type,
            (case when nullable = 'N' then 1 else 0 end) as not_null,
            (case when data_default is not null then 1 else 0 end) as has_default
        FROM all_tab_columns
        WHERE LOWER(table_name) = LOWER(:1)
        AND REGEXP_LIKE(LOWER(owner), '^' || COALESCE(LOWER(:2), '.*')  || '$')
        """).strip()

    def __init__(self):
        super().__init__()
        self.required_params = {'host', 'port', 'dbname', 'user'}
        self.missing_driver_msg = (
            "Could not import oracledb module required for Oracle connections.  "
            "See https://github.com/BritishGeologicalSurvey/etlhelper for installation instructions")
        self.named_paramstyle = 'named'
        self.positional_paramstyle = 'numeric'

        try:
            import oracledb
            self.sql_exceptions = (oracledb.DatabaseError,
                                   oracledb.InterfaceError)
            self.connect_exceptions = (oracledb.DatabaseError,
                                       oracledb.InterfaceError)
            self.paramstyle = oracledb.paramstyle
            self._connect_func = oracledb.connect

            # Fetch large objects as Python strings/bytes instead of LOBs
            # The oracledb default is LOB, but native types are faster to
            # transfer and are in line with return types of other databases
            # e.g. SQLite or PostgreSQL.
            # https://python-oracledb.readthedocs.io/en/latest/user_guide/lob_data.html#fetching-lobs-as-strings-and-bytes
            oracledb.defaults.fetch_lobs = False
        except ImportError:
            warnings.warn(self.missing_driver_msg)

    def get_connection_string(self, db_params, password_variable):
        """
        Return a connection string
        :param db_params: DbParams
        :param password_variable: str, password
        :return: str
        """
        # Prepare connection string
        password = self.get_password(password_variable)
        return (f'{db_params.user}/{password}@'
                f'{db_params.host}:{db_params.port}/{db_params.dbname}')

    def get_sqlalchemy_connection_string(self, db_params, password_variable):
        """
        Returns connection string for SQLAlchemy engine.
        """
        password = self.get_password(password_variable)
        return (f'oracle+oracledb://{db_params.user}:{password}@'
                f'{db_params.host}:{db_params.port}/{db_params.dbname}')



"""
Functions used to handle aborts during threaded operations.
"""
import threading

from etlhelper.exceptions import ETLHelperAbort

abort_event = threading.Event()


def abort_etlhelper_threads():
    """
    Abort the ETL Helper process at the end of the current chunk.  During
    threaded operation, this will cause an ETL HelperAbort error to be
    raised.
    """
    abort_event.set()


def clear_abort_event():
    """Clear abort_event."""
    abort_event.clear()


def raise_for_abort(message):
    """Raise EtlHelperAbort exception with message if abort_event is set."""
    if abort_event.is_set():
        raise ETLHelperAbort(message)



"""ETL Helper script to demonstrate copying records between databases with iter_rows and load."""
import etlhelper as etl
from my_databases import POSTGRESDB, ORACLEDB

select_sql = """
    SELECT id, name, value FROM my_table
    WHERE value > :min_value
"""

with ORACLEDB.connect("ORA_PASSWORD") as src_conn:
    with POSTGRESDB.connect("PG_PASSWORD") as dest_conn:
        rows = etl.iter_rows(select_sql, src_conn, parameters={"min_value": 99})
        etl.load("my_table", dest_conn, rows)



"""Tests for db_params that require a database connection."""
import pytest

from etlhelper import DbParams

# pylint: disable=unused-argument, missing-docstring


def test_is_reachable(pgtestdb_params):
    # Use pgtestdb_params here as it changes hostname for CI tests
    assert pgtestdb_params.is_reachable()


def test_is_unreachable(pgtestdb_params):
    # Use pgtestdb_params here as it changes hostname for CI tests
    dbparam = pgtestdb_params.copy()  # Copy so real PGTESTDB is not modified
    dbparam['port'] = 1
    assert dbparam.is_reachable() is False


def test_sqlite_dbparam_not_supported():
    sqlitedb = DbParams(
        dbtype='SQLITE',
        filename='sqlite.db')

    with pytest.raises(ValueError):
        sqlitedb.is_reachable()



"""
Database helper for SQLite
"""
import warnings
from contextlib import contextmanager
from textwrap import dedent

from etlhelper.db_helpers.db_helper import DbHelper


class SQLiteDbHelper(DbHelper):
    """
    SQLite DB helper class
    """
    # schema_name is not used for SQLite but is required as parameter to be
    # consistent with other databases.  The WHERE clause is always true,
    # whether schema_name is NULL or not.
    table_info_query = dedent("""
        SELECT
            name,
            lower(type),
            "notnull" as not_null,
            (case when dflt_value is not null then 1 else 0 end) as has_default
        FROM pragma_table_info(:table_name)
        -- this effectively ignores the unused schema_name
        -- parameter since schemas are not used in sqlite
        WHERE COALESCE(TRUE, :schema_name)
        ;""").strip()

    def __init__(self):
        super().__init__()
        self.required_params = {'filename'}
        self.missing_driver_msg = (
            "Could not import sqlite3 module required for SQLite connections.  "
            "Check Python configuration - this should be part of Standard Library.")
        self.named_paramstyle = 'named'
        self.positional_paramstyle = 'qmark'

        try:
            import sqlite3
            self.sql_exceptions = (sqlite3.DatabaseError,
                                   sqlite3.InterfaceError)
            self.connect_exceptions = (sqlite3.DatabaseError,
                                       sqlite3.InterfaceError)
            self.paramstyle = sqlite3.paramstyle
            self._connect_func = sqlite3.connect
        except ImportError:
            warnings.warn(self.missing_driver_msg)

    def get_connection_string(self, db_params, password_variable=None):
        """
        Return a connection string
        :param db_params: DbParams
        :return: str
        """
        # Prepare connection string
        # Accept unused password_variable for consistency with other databases
        return (f'{db_params.filename}')

    def get_sqlalchemy_connection_string(self, db_params,
                                         password_variable=None):
        """
        Returns connection string for SQLAlchemy type connections
        :param db_params: DbParams
        :return: str
        """
        return (f'sqlite:///{db_params.filename}')

    @staticmethod
    @contextmanager
    def cursor(conn):
        """
        Return a cursor on current connection.  This implementation allows
        SQLite cursor to be used as context manager as with other db types.
        """
        try:
            cursor = conn.cursor()
            yield cursor
        finally:
            cursor.close()



"""ETL Helper script to demonstrate using a transform function which yields individual rows."""
import sqlite3
from typing import Iterator
import etlhelper as etl
from etlhelper.row_factories import dict_row_factory

db_file = "igneous_rocks.db"
select_sql = "SELECT * FROM igneous_rock"


def my_transform(chunk: Iterator[dict]) -> Iterator[dict]:
    # Add prefix to id, remove newlines, set lower case names

    for row in chunk:  # each row is a dictionary (mutable)
        row["id"] += 1000
        row["description"] = row["description"].replace("\n", " ")
        row["name"] = row["name"].lower()
        yield row


with sqlite3.connect(db_file) as conn:
    rows = etl.fetchall(
        select_sql,
        conn,
        row_factory=dict_row_factory,
        transform=my_transform,
    )



"""ETL Helper script to demonstrate using fetch functions."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"

with sqlite3.connect(db_file) as conn:
    first_row = etl.fetchone("SELECT * FROM igneous_rock", conn)
    all_rows = etl.fetchall("SELECT * FROM igneous_rock", conn)

print(first_row)
print(all_rows)



"""Tests for utils functions."""
import pytest

from etlhelper.exceptions import ETLHelperQueryError
from etlhelper.utils import (
    Column,
    table_info,
)

# pylint: disable=unused-argument, missing-docstring


def test_table_info_no_schema_no_duplicates(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange
    expected = [
        Column('id', 'integer', not_null=1, has_default=0),
        Column('value', 'double precision', not_null=1, has_default=0),
        Column('simple_text', 'text', not_null=0, has_default=1),
        Column('utf8_text', 'text', not_null=0, has_default=0),
        Column('day', 'date', not_null=0, has_default=0),
        Column('date_time', 'timestamp without time zone', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', pgtestdb_conn)

    # Assert
    assert columns == expected


def test_table_info_with_schema_no_duplicates(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange
    expected = [
        Column('id', 'integer', not_null=1, has_default=0),
        Column('value', 'double precision', not_null=1, has_default=0),
        Column('simple_text', 'text', not_null=0, has_default=1),
        Column('utf8_text', 'text', not_null=0, has_default=0),
        Column('day', 'date', not_null=0, has_default=0),
        Column('date_time', 'timestamp without time zone', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', pgtestdb_conn, schema='public')

    # Assert
    assert columns == expected


def test_table_info_no_schema_with_duplicates(pgtestdb_conn, duplicate_schema):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r'Table name src is not unique'):
        table_info('src', pgtestdb_conn)


def test_table_info_with_schema_with_duplicates(pgtestdb_conn, duplicate_schema):
    # Arrange
    expected = [
        Column('id', 'integer', not_null=1, has_default=0),
        Column('value', 'double precision', not_null=1, has_default=0),
        Column('simple_text', 'text', not_null=0, has_default=1),
        Column('utf8_text', 'text', not_null=0, has_default=0),
        Column('day', 'date', not_null=0, has_default=0),
        Column('date_time', 'timestamp without time zone', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', pgtestdb_conn, schema='public')

    # Assert
    assert columns == expected


def test_table_info_bad_table_name_no_schema(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'bad_table' not found."):
        table_info('bad_table', pgtestdb_conn)


def test_table_info_bad_table_name_with_schema(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'public.bad_table' not found."):
        table_info('bad_table', pgtestdb_conn, schema='public')


# Fixtures here

@pytest.fixture(scope='function')
def duplicate_schema(pgtestdb_conn, pgtestdb_test_tables):
    # Set up
    with pgtestdb_conn.cursor() as cursor:
        # Create a duplicate of the test tables in a new schema
        cursor.execute("CREATE SCHEMA IF NOT EXISTS duplicate", pgtestdb_conn)
        cursor.execute("SELECT * INTO duplicate.src FROM src", pgtestdb_conn)
    pgtestdb_conn.commit()

    # Return control to run test
    yield

    # Tear down
    with pgtestdb_conn.cursor() as cursor:
        cursor.execute("DROP SCHEMA IF EXISTS duplicate CASCADE", pgtestdb_conn)
    pgtestdb_conn.commit()



"""ETL Helper script to demonstrate using a transform function which returns a list of rows."""
import random
import sqlite3
from typing import Iterator
import etlhelper as etl
from etlhelper.row_factories import namedtuple_row_factory

db_file = "igneous_rocks.db"
select_sql = "SELECT * FROM igneous_rock"


def my_transform(chunk: Iterator[tuple]) -> list[tuple]:
    # Append random integer (1-10), filter if <5.

    new_chunk = []
    for row in chunk:  # each row is a namedtuple (immutable)
        extra_value = random.randrange(10)
        if extra_value >= 5:  # some rows are dropped
            new_row = (*row, extra_value)  # new rows have extra column
            new_chunk.append(new_row)

    return new_chunk


with sqlite3.connect(db_file) as conn:
    rows = etl.fetchall(
        select_sql,
        conn,
        row_factory=namedtuple_row_factory,
        transform=my_transform,
    )



"""Tests for etl functions"""
import pytest

from etlhelper.etl import validate_identifier
from etlhelper.exceptions import ETLHelperBadIdentifierError


@pytest.mark.parametrize('good_identifier', [
    "all_alpha",
    "ALL_CAPS_ALPHA",
    "þis_is_nøn_latîn_ælpha",
    "this_is_nøn_latîn_ælpha_except_first_character",
    "_starts_with_underscore",
    "has_numbers_123",
    "has_$",
    # schema qualified table names with same rules
    "schema.all_alpha",
    "schema.ALL_CAPS_ALPHA",
    "schema_123.þis_is_nøn_latîn_ælpha",
    "schema.this_is_nøn_latîn_ælpha_except_first_character",
    "schema._starts_with_underscore",
    "schema.has_numbers_123",
    "schema.has_$",
    # rules also apply to schema names
    "all_alpha.table",
    "ALL_CAPS_ALPHA.table",
    "þis_is_nøn_latîn_ælpha.table",
    "this_is_nøn_latîn_ælpha_except_first_character.table",
    "_starts_with_underscore.table",
    "has_numbers_123.table",
    "has_$.table",
])
def test_validate_identifier_good(good_identifier):
    validate_identifier(good_identifier)


@pytest.mark.parametrize('bad_identifier', [
    "Robert'); DROP TABLE students; --",  # SQL injection attempt: https://xkcd.com/327/
    "",  # empty
    "123_starts_number",
    "$_starts_dollar",
    ";",
    "()",
    "schema.123_starts_number",
    "schema.$_starts_dollar",
    "123_starts_number.table",
    "$_starts_dollar.table",
    ";",
    "()",
])
def test_validate_identifier_bad(bad_identifier):
    with pytest.raises(ETLHelperBadIdentifierError):
        validate_identifier(bad_identifier)



"""ETL Helper script to demonstrate copying data from an Oracle database into a PostgreSQL database."""
import datetime as dt
from textwrap import dedent
import etlhelper as etl
from my_databases import ORACLEDB, POSTGRESDB

CREATE_SQL = dedent("""
    CREATE TABLE IF NOT EXISTS sensordata.readings
    (
        sensor_data_id bigint PRIMARY KEY,
        measure_id bigint,
        time_stamp timestamp without time zone,
        meas_value double precision
    )
    """).strip()

DELETE_SQL = dedent("""
    DELETE FROM sensordata.readings
    WHERE time_stamp BETWEEN %(startdate)s AND %(enddate)s
    """).strip()

SELECT_SQL = dedent("""
    SELECT id, measure_id, time_stamp, reading
    FROM sensor_data
    WHERE time_stamp BETWEEN :startdate AND :enddate
    ORDER BY time_stamp
    """).strip()

INSERT_SQL = dedent("""
    INSERT INTO sensordata.readings (sensor_data_id, measure_id, time_stamp,
        meas_value)
    VALUES (%s, %s, %s, %s)
    """).strip()


def copy_readings(startdate: dt.datetime, enddate: dt.datetime) -> None:
    params = {"startdate": startdate, "enddate": enddate}

    with ORACLEDB.connect("ORA_PASSWORD") as src_conn:
        with POSTGRESDB.connect("PG_PASSWORD") as dest_conn:
            etl.execute(CREATE_SQL, dest_conn)
            etl.execute(DELETE_SQL, dest_conn, parameters=params)
            etl.copy_rows(
                SELECT_SQL,
                src_conn,
                INSERT_SQL,
                dest_conn,
                parameters=params,
            )


if __name__ == "__main__":
    # Copy data from 00:00:00 yesterday to 00:00:00 today
    today = dt.combine(dt.date.today(), dt.time.min)
    yesterday = today - dt.timedelta(1)

    copy_readings(yesterday, today)
